from kome.insert import InsertClause
from kome.select import SelectClause
from kome.update import UpdateClause


class Query:
    @staticmethod
    def select():
        return SelectClause()

    @staticmethod
    def insert():
        return InsertClause()

    @staticmethod
    def update():
        return UpdateClause()



from typing import Union

from kome.const import Comparator, Equality
from kome.source import QueryObject
from kome.utils import format_quotes

Premitive = Union[str, int, float, bool]
PyExpression = Union[Premitive, "BasicExpression"]


def force_expression(value: PyExpression):
    if isinstance(value, BasicExpression):
        return value
    if isinstance(value, (str, int, bool, float)):
        return PyPremitiveExpression(value)
    raise Exception(f"not support, {value}")


class BasicExpression(QueryObject):
    def as_(self, alias: str):
        return AliasedExpression(self, alias)

    def __eq__(self, o: PyExpression):
        return Condition(self, force_expression(o), Equality.eq)

    def __ne__(self, o: PyExpression):
        return Condition(self, force_expression(o), Equality.ne)

    def __gt__(self, o: PyExpression):
        return Condition(self, force_expression(o), Equality.gt)

    def __gte__(self, o: PyExpression):
        return Condition(self, force_expression(o), Equality.gte)

    def __lt__(self, o: PyExpression):
        return Condition(self, force_expression(o), Equality.lt)

    def __lte__(self, o: PyExpression):
        return Condition(self, force_expression(o), Equality.lte)


class PyPremitiveExpression(BasicExpression):
    def __init__(self, value: Premitive) -> None:
        self.value = value

    @property
    def sql(self) -> str:
        if isinstance(self.value, bool):
            return "TRUE" if self.value else "FALSE"
        if isinstance(self.value, str):
            return f"'{self.value}'"
        if isinstance(self.value, int):
            return str(self.value)
        raise Exception(f"wait for support. {type(self.value)}")


class AliasedExpression(QueryObject):
    def __init__(self, expr: BasicExpression, alias: str) -> None:
        self.expr = expr
        self.alias = alias

    @property
    def sql(self) -> str:
        return f"{self.expr.sql} AS {format_quotes(self.alias)}"


Expression = Union[BasicExpression, AliasedExpression]


class Condition(BasicExpression):
    def __init__(
        self, left: BasicExpression, right: BasicExpression, comparator: Comparator
    ) -> None:
        self.left = left
        self.right = right
        self.comparator = comparator

    @property
    def sql(self) -> str:
        return f"{self.left.sql}{self.comparator.value}{self.right.sql}"


class P(BasicExpression):
    def __init__(self, placeholder: str) -> None:
        self.placeholder = placeholder

    @property
    def sql(self) -> str:
        return str(self.placeholder)



from typing import List, Optional, Sequence

from kome.const import JoinType, OrderType
from kome.expression import Condition, Expression
from kome.source import QueryObject, QuerySource


class Order:
    def __init__(self, expr: Expression, order: OrderType) -> None:
        self.expr = expr
        self.order = order

    def __str__(self) -> str:
        return f"{self.expr.sql} {self.order.value}"


class Join:
    def __init__(self, join_type: JoinType, source: QuerySource, on: Condition) -> None:
        self.join_type = join_type
        self.source = source
        self.on = on

    def __str__(self) -> str:
        return f"{self.join_type.value} JOIN ({self.source.sql}) ON {self.on.sql}"


class Select:
    def __init__(self) -> None:
        self.distinct_ = False
        self.select: List[Expression] = []
        self.source: Optional[QuerySource] = None
        self.join: List[Join] = []
        self.where: Optional[Condition] = None
        self.group_by: List[Expression] = []
        self.having: Optional[Condition] = None
        self.order: List[Order] = []
        self.limit: Optional[int] = None

    @property
    def sql(self):
        sql = "SELECT "
        if self.distinct_:
            sql += "DISTINCT "
        sql += ",".join(map(lambda s: s.sql, self.select))
        if not self.source:
            raise Exception("Select Query shoud call 'from_' method.")
        sql += f" FROM ({self.source.sql})"
        if self.join:
            sql += " "
            sql += " ".join(map(str, self.join))
        if self.where:
            sql += f" WHERE {self.where.sql}"
        if self.group_by:
            sql += " GROUP BY "
            sql += ",".join(map(lambda s: s.sql, self.group_by))
        if self.having:
            sql += f" HAVING {self.having.sql}"
        if self.order:
            sql += " ORDER BY "
            sql += ",".join(map(str, self.order))
        if self.limit:
            sql += f" LIMIT {self.limit}"
        return sql


class SelectMixin(QueryObject):
    _select: Select

    @property
    def query(self):
        return self._select

    @property
    def sql(self) -> str:
        return self._select.sql


class SelectClause:
    def __init__(self):
        self._select = Select()

    def from_(self, source: QuerySource):
        return SelectFrom(self._select, source)


class SelectFrom:
    def __init__(self, select: Select, source: QuerySource) -> None:
        select.source = source
        self._select = select

    def select(self, column: Expression, *columns: Expression):
        return SelectTarget(self._select, column, *columns)


class SelectTarget(SelectMixin):
    def __init__(
        self, select: Select, column: Expression, *columns: Expression
    ) -> None:
        select.select.append(column)
        select.select.extend(columns)
        self._select = select

    def distinct(self):
        return SelectDistinct(self._select)

    def left_join(
        self,
        source: QuerySource,
        on: Condition,
        selects: Optional[Sequence[Expression]] = None,
    ):
        return SelectJoin(self._select, JoinType.left, source, on, selects)

    def inner_join(
        self,
        source: QuerySource,
        on: Condition,
        selects: Optional[Sequence[Expression]] = None,
    ):
        return SelectJoin(self._select, JoinType.inner, source, on, selects)

    def where(self, condition: Condition):
        return SelectWhere(self._select, condition)

    def group_by(self, expr: Expression, *exprs: Expression):
        return SelectGroupBy(self._select, expr, *exprs)

    def order_by(self, expr: Expression, order: OrderType = OrderType.asc):
        return SelectOrderBy(self._select, expr, order)

    def limit(self, limit: int):
        return SelectLimit(self._select, limit)


class SelectDistinct(SelectMixin):
    def __init__(self, select: Select) -> None:
        select.distinct_ = True
        self._select = select

    def left_join(
        self,
        source: QuerySource,
        on: Condition,
        selects: Optional[Sequence[Expression]] = None,
    ):
        return SelectJoin(self._select, JoinType.left, source, on, selects)

    def inner_join(
        self,
        source: QuerySource,
        on: Condition,
        selects: Optional[Sequence[Expression]] = None,
    ):
        return SelectJoin(self._select, JoinType.inner, source, on, selects)

    def where(self, condition: Condition):
        return SelectWhere(self._select, condition)

    def group_by(self, expr: Expression, *exprs: Expression):
        return SelectGroupBy(self._select, expr, *exprs)

    def order_by(self, expr: Expression, order: OrderType = OrderType.asc):
        return SelectOrderBy(self._select, expr, order)

    def limit(self, limit: int):
        return SelectLimit(self._select, limit)


class SelectJoin(SelectMixin):
    def __init__(
        self,
        select: Select,
        join_type: JoinType,
        source: QuerySource,
        on: Condition,
        selects: Optional[Sequence[Expression]] = None,
    ):
        select.join.append(Join(join_type, source, on))
        if selects:
            select.select.extend(selects)
        self._select = select

    def left_join(
        self,
        source: QuerySource,
        on: Condition,
        selects: Optional[Sequence[Expression]] = None,
    ):
        return SelectJoin(self._select, JoinType.left, source, on, selects)

    def inner_join(
        self,
        source: QuerySource,
        on: Condition,
        selects: Optional[Sequence[Expression]] = None,
    ):
        return SelectJoin(self._select, JoinType.inner, source, on, selects)

    def where(self, condition: Condition):
        return SelectWhere(self._select, condition)

    def group_by(self, expr: Expression, *exprs: Expression):
        return SelectGroupBy(self._select, expr, *exprs)

    def order_by(self, expr: Expression, order: OrderType = OrderType.asc):
        return SelectOrderBy(self._select, expr, order)

    def limit(self, limit: int):
        return SelectLimit(self._select, limit)


class SelectWhere(SelectMixin):
    def __init__(self, select: Select, condition: Condition):
        select.where = condition
        self._select = select

    def group_by(self, expr: Expression, *exprs: Expression):
        return SelectGroupBy(self._select, expr, *exprs)

    def order_by(self, expr: Expression, order: OrderType = OrderType.asc):
        return SelectOrderBy(self._select, expr, order)

    def limit(self, limit: int):
        return SelectLimit(self._select, limit)


class SelectGroupBy(SelectMixin):
    def __init__(self, select: Select, expr: Expression, *exprs: Expression) -> None:
        select.group_by.append(expr)
        select.group_by.extend(exprs)
        self._select = select

    def having(self, condition: Condition):
        return SelectHaving(self._select, condition)

    def order_by(self, expr: Expression, order: OrderType = OrderType.asc):
        return SelectOrderBy(self._select, expr, order)

    def limit(self, limit: int):
        return SelectLimit(self._select, limit)


class SelectHaving(SelectMixin):
    def __init__(self, select: Select, condition: Condition) -> None:
        select.having = condition
        self._select = select

    def order_by(self, expr: Expression, order: OrderType = OrderType.asc):
        return SelectOrderBy(self._select, expr, order)

    def limit(self, limit: int):
        return SelectLimit(self._select, limit)


class SelectOrderBy(SelectMixin):
    def __init__(self, select: Select, expr: Expression, order: OrderType):
        select.order.append(Order(expr, order))
        self._select = select

    def order_by(self, expr: Expression, order: OrderType = OrderType.asc):
        return SelectOrderBy(self._select, expr, order)

    def limit(self, limit: int):
        return SelectLimit(self._select, limit)


class SelectLimit(SelectMixin):
    def __init__(self, select: Select, limit: int) -> None:
        if limit < 1:
            raise SelectLimit.SelectLimitError(limit)
        select.limit = limit
        self._select = select

    class SelectLimitError(Exception):
        def __init__(self, limit: int) -> None:
            super().__init__(f"Limt must be positive integer. But got {limit}")



from typing import Union

from kome.source import AliasedQuerySource, BasicQuerySource
from kome.utils import format_quotes


class BasicTable(BasicQuerySource):
    def __init__(self, table_name: str) -> None:
        self.table_name = table_name

    def as_(self, alias: str):
        return AliasedTable(self, alias)

    @property
    def sql(self) -> str:
        return f"{format_quotes(self.table_name)}"


class AliasedTable(AliasedQuerySource):
    def __init__(self, table: BasicTable, alias: str) -> None:
        self.table = table
        self.alias = alias

    @property
    def sql(self) -> str:
        return f"{self.table.sql} AS {format_quotes(self.alias)}"


Table = Union[BasicTable, AliasedTable]



from typing import Union

from kome.utils import format_quotes


class QueryObject:
    @property
    def sql(self) -> str:
        raise NotImplementedError()


class BasicQuerySource(QueryObject):
    def as_(self, alias: str) -> QueryObject:
        return AliasedQuerySource(self, alias)


class AliasedQuerySource(QueryObject):
    def __init__(self, qs: BasicQuerySource, alias: str) -> None:
        self.qs = qs
        self.alias = alias

    @property
    def sql(self) -> str:
        return f"{self.qs.sql} AS {format_quotes(self.alias)}"


QuerySource = Union[BasicQuerySource, AliasedQuerySource]



from typing import List, Optional, Union

from kome.expression import BasicExpression, P, Premitive, force_expression
from kome.field import Field
from kome.source import QueryObject
from kome.table import BasicTable
from kome.utils import format_quotes

InsertableValue = Union[Premitive, P]
# query
"""
INSERT INTO テーブル名 [ (列名1 [ ,列名2・・・]) ]
VALUES (値A1 [, 値A2 ...]) [, (値B1 [, 値B2 ...]) ...];
"""


class Insert:
    def __init__(self) -> None:
        self.table: Optional[BasicTable] = None
        self.fields: List[Field] = []
        self.values: List[List[BasicExpression]] = []

    @property
    def sql(self) -> str:
        if self.table is None:
            raise RuntimeError("Insert Query shoud call 'into'")
        sql = f"INSERT INTO {self.table.sql} ({','.join(map(lambda v:format_quotes(v.field_name) ,self.fields))}) VALUES "
        sql += ",".join(
            map(lambda v: "(" + ",".join(map(lambda k: k.sql, v)) + ")", self.values)
        )
        return sql


class InsertMixin(QueryObject):
    _insert: Insert

    @property
    def query(self):
        return self._insert

    @property
    def sql(self) -> str:
        return self._insert.sql


class InsertClause:
    def __init__(self) -> None:
        self._insert = Insert()

    def into(self, table: BasicTable):
        return InsertInto(self._insert, table)


class InsertInto:
    def __init__(self, insert: Insert, table: BasicTable) -> None:
        insert.table = table
        self._insert = insert

    def columns(self, field: Field, *fields: Field):
        return InsertColumns(self._insert, field, *fields)


class InsertColumns:
    def __init__(self, insert: Insert, field: Field, *fields: Field) -> None:
        insert.fields.append(field)
        insert.fields.extend(fields)
        self._insert = insert

    def values(self, value: InsertableValue, *values: InsertableValue):
        return InsertValues(self._insert, value, *values)


class InsertValues(InsertMixin):
    def __init__(
        self, insert: Insert, value: InsertableValue, *values: InsertableValue
    ) -> None:
        if len(insert.fields) != 1 + len(values):
            raise ValueError(
                f"inser value length not match. \n\t{insert.fields=}\n\tvalues={(value,*values)}"
            )
        insert.values.append(list(map(force_expression, [value, *values])))
        self._insert = insert

    def values(self, value: InsertableValue, *values: InsertableValue):
        return InsertValues(self._insert, value, *values)



from kome.expression import BasicExpression


class SQLFunction(BasicExpression):
    def __init__(self, func_name: str, *args: BasicExpression) -> None:
        self.func_name = func_name
        self.args = args

    @property
    def sql(self) -> str:
        return f"{self.func_name}({','.join(map(lambda v:v.sql,self.args))})"


def count(expr: BasicExpression):
    return SQLFunction("COUNT", expr)



from typing import List, Optional

from kome.expression import BasicExpression, Condition, PyExpression, force_expression
from kome.field import Field
from kome.source import QueryObject
from kome.table import Table


class UpdateSet:
    def __init__(self, field: Field, expr: BasicExpression) -> None:
        self.field = field
        self.expr = expr

    def __str__(self) -> str:
        return f"{self.field.sql}={self.expr.sql}"


class Update:
    def __init__(self) -> None:
        self.table: Optional[Table] = None
        self.sets: List[UpdateSet] = []
        self.where: Optional[Condition] = None

    @property
    def sql(self):
        if not self.table:
            raise RuntimeError("Update Query shoud call 'from_'")
        sql = f"UPDATE {self.table.sql} SET {','.join(map(str, self.sets))}"
        if self.where:
            sql += f" WHERE {self.where.sql}"
        return sql


class UpdateMixin(QueryObject):
    _update: Update

    @property
    def query(self):
        return self._update

    @property
    def sql(self) -> str:
        return self._update.sql


class UpdateClause:
    def __init__(self) -> None:
        self._update = Update()

    def from_(self, table: Table):
        return UpdateFrom(self._update, table)


class UpdateFrom:
    def __init__(self, update: Update, table: Table) -> None:
        update.table = table
        self._update = update

    def set(self, field: Field, expr: PyExpression):
        return UpdateValue(self._update, field, expr)


class UpdateValue(UpdateMixin):
    def __init__(self, update: Update, field: Field, expr: PyExpression) -> None:
        update.sets.append(UpdateSet(field, force_expression(expr)))
        self._update = update

    def set(self, field: Field, expr: PyExpression):
        return UpdateValue(self._update, field, expr)

    def where(self, condition: Condition):
        return UpdateWhere(self._update, condition)


class UpdateWhere(UpdateMixin):
    def __init__(self, update: Update, condition: Condition) -> None:
        update.where = condition
        self._update = update



from typing import Optional

from kome.expression import BasicExpression
from kome.table import AliasedTable, Table
from kome.utils import format_quotes


class Field(BasicExpression):
    def __init__(self, field_name: str, table: Optional[Table] = None):
        self.field_name = field_name
        self.table = table

    @property
    def sql(self) -> str:
        if self.table is None:
            return format_quotes(self.field_name)
        if isinstance(self.table, AliasedTable):
            return f"{format_quotes(self.table.alias)}.{format_quotes(self.field_name)}"
        return f"{self.table.sql}.{format_quotes(self.field_name)}"



from enum import Enum


class OrderType(Enum):
    desc = "DESC"
    asc = "ASC"


class Arithmetic(Enum):
    add = "+"
    sub = "-"
    mul = "*"
    div = "/"


class Comparator(Enum):
    pass


class Equality(Comparator):
    eq = "="
    ne = "<>"
    gt = ">"
    gte = ">="
    lt = "<"
    lte = "<="


class Matching(Comparator):
    not_like = " NOT LIKE "
    like = " LIKE "
    not_ilike = " NOT ILIKE "
    ilike = " ILIKE "
    regex = " REGEX "
    bin_regex = " REGEX BINARY "


class JoinType(Enum):
    left = "LEFT"
    inner = "INNER"



from typing import Optional

from kome.expression import Condition
from kome.source import QueryObject
from kome.table import BasicTable, Table
from kome.utils import format_quotes


class Delete:
    def __init__(self) -> None:
        self.table: Optional[Table] = None
        self.where: Optional[Condition] = None

    @property
    def sql(self) -> str:
        if self.table is None:
            raise RuntimeError("Delete Query shoud call 'from_'")
        if isinstance(self.table, BasicTable):
            sql = f"DELETE FROM {self.table.sql}"
        else:
            sql = f"DELETE {format_quotes(self.table.alias)} FROM {self.table.sql}"
        if self.where:
            sql += f" WHERE {self.where.sql}"
        return sql


class DeleteMixin(QueryObject):
    _delete: Delete

    @property
    def query(self):
        return self._delete

    @property
    def sql(self) -> str:
        return self._delete.sql


class DeleteClause:
    def __init__(self) -> None:
        self._delete = Delete()

    def from_(self, table: Table):
        return DeleteFrom(self._delete, table)


class DeleteFrom:
    def __init__(self, delete: Delete, table: Table) -> None:
        delete.table = table
        self._delete = delete

    def where(self, condition: Condition):
        return DeleteWhere(self._delete, condition)

    def all(self):
        # delete all is too dangerous.
        # So if you need to delete all records, you should expressly say all.
        return DeleteAll(self._delete)


class DeleteAll(DeleteMixin):
    def __init__(self, delete: Delete) -> None:
        self._delete = delete


class DeleteWhere(DeleteMixin):
    def __init__(self, delete: Delete, condition: Condition) -> None:
        delete.where = condition
        self._delete = delete
